/*  Pcsx - Pc Psx Emulator
 *  Copyright (C) 1999-2002  Pcsx Team
 *  Copyright (C) 2008 Andrew Burton
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "Config.h"
#include "Linux.h"

#define PCSX_DOT_DIR "/.pcsx/"	/* TODO - Same as in Gtk2Gui.c */

#define READONLY "r"
#define READWRITE "w"


gboolean g_key_file_save_to_file(GKeyFile *kf, const gchar *file, GError **error);
static void pcsx_gui_prefs_set_int_property_to_key_file (PcsxGuiPrefs *pr, GParamSpec *param, gpointer user_data);
static void pcsx_gui_prefs_set_string_property_to_key_file (PcsxGuiPrefs *pr, GParamSpec *param, gpointer user_data);

static gint pcsx_gui_prefs_get_int_property_from_key_file (PcsxGuiPrefs *pr, gchar *property);
static gchar* pcsx_gui_prefs_get_string_property_from_key_file (PcsxGuiPrefs *pr, gchar *property);

static void pcsx_gui_prefs_class_init (PcsxGuiPrefsClass *klass);
static void pcsx_gui_prefs_init (PcsxGuiPrefs *pr);
static void pcsx_gui_prefs_finalize (GObject *obj);
static void pcsx_gui_prefs_set_property (GObject *object,
					 guint prop_id,
					 const GValue *value,
					 GParamSpec *pspec);
static void pcsx_gui_prefs_get_property (GObject *object,
					 guint prop_id,
					 GValue *value,
					 GParamSpec *pspec);

G_DEFINE_TYPE (PcsxGuiPrefs, pcsx_gui_prefs, G_TYPE_OBJECT)

struct _PcsxGuiPrefsPrivate {

	/* References to the ini file that contains our prefs */
	GKeyFile *prefs_ini_file;
	gchar *filename;
	
	gchar *biosdir;
	gchar *plugindir;

	gchar *bios;
	gint loadcdbios;
	
	gchar *gpu_plugin;
	gchar *spu_plugin;
	gchar *cdr_plugin;
	gchar *pad1_plugin;
	gchar *pad2_plugin;
	
	/* CPU-related properties */
	gint xa;
	gint sio;
	gint mdec;
	gint psxauto;
	gint cdda;
	gint cpu;
	gint dbg;
	gint psxout;
	gint psxtype;
	gint spuirq;
	gint rcntfix;
	gint vsyncwa;
	
	gchar *currentfile;
	
	/* Memory cards */
	gchar *memcard1;
	gchar *memcard2;
	
};

/*
 Called by changes to the pcsx_gui_prefs GObject; we also need to update the
 Config struct for some of the properties, since that is referenced in
 libpcsxcore rather than the GObject
 */
static void
pcsx_gui_prefs_set_property (GObject *object,
			     guint prop_id,
			     const GValue *value,
			     GParamSpec *pspec)
{
	PcsxGuiPrefs *prefs;
	
	prefs = PCSX_GUI_PREFS (object);

	switch (prop_id) {
		
		case PROP_GUI_PREFS_DIR_BIOS:
			prefs->priv->biosdir = g_strdup (g_value_get_string (value));
			g_snprintf(Config.BiosDir, MAXPATHLEN, "%s", prefs->priv->biosdir);
			break;
		case PROP_GUI_PREFS_DIR_PLUGINS:
			prefs->priv->plugindir = g_strdup (g_value_get_string (value));
			g_snprintf(Config.PluginsDir, MAXPATHLEN, "%s", prefs->priv->bios);
			break;
		case PROP_GUI_PREFS_BIOS:
			prefs->priv->bios = g_strdup (g_value_get_string (value));
			g_snprintf(Config.Bios, 255, "%s", prefs->priv->bios);
			break;
		case PROP_GUI_PREFS_LOADCDBIOS:
			prefs->priv->loadcdbios = g_value_get_int (value);
			break;
		case PROP_GUI_PREFS_PLUGIN_GPU:
			prefs->priv->gpu_plugin = g_strdup (g_value_get_string (value));
			g_snprintf(Config.Gpu, 255, "%s", prefs->priv->gpu_plugin);
			break;
		case PROP_GUI_PREFS_PLUGIN_SPU:
			prefs->priv->spu_plugin = g_strdup (g_value_get_string (value));
			g_snprintf(Config.Spu, 255, "%s", prefs->priv->spu_plugin);
			break;
		case PROP_GUI_PREFS_PLUGIN_PAD1:
			prefs->priv->pad1_plugin = g_strdup (g_value_get_string (value));
			g_snprintf(Config.Pad1, 255, "%s", prefs->priv->pad1_plugin);
			break;
		 case PROP_GUI_PREFS_PLUGIN_PAD2:
			prefs->priv->pad2_plugin = g_strdup (g_value_get_string (value));
			g_snprintf(Config.Pad2, 255, "%s", prefs->priv->pad2_plugin);
			break;
		case PROP_GUI_PREFS_PLUGIN_CDR:
			prefs->priv->cdr_plugin = g_strdup (g_value_get_string (value));
			g_snprintf(Config.Cdr, 255, "%s", prefs->priv->cdr_plugin);
			break;
		case PROP_GUI_PREFS_CPU_XA:
			prefs->priv->xa = g_value_get_int (value);
			Config.Xa = prefs->priv->xa;
			break;
		case PROP_GUI_PREFS_CPU_SIO:
			prefs->priv->sio = g_value_get_int (value);
			Config.Sio = prefs->priv->sio;
			break;
		case PROP_GUI_PREFS_CPU_MDEC:
			prefs->priv->mdec = g_value_get_int (value);
			Config.Mdec = prefs->priv->mdec;
			break;
		case PROP_GUI_PREFS_CPU_PSXTYPE:
			prefs->priv->psxtype = g_value_get_int (value);
			Config.PsxType = prefs->priv->psxtype;
			break;
		case PROP_GUI_PREFS_CPU_PSXAUTO:
			prefs->priv->psxauto = g_value_get_int (value);
			Config.PsxAuto = prefs->priv->psxauto;
			break;
		case PROP_GUI_PREFS_CPU_CDDA:		
			prefs->priv->cdda = g_value_get_int (value);
			Config.Cdda = prefs->priv->cdda;
			break;
		case PROP_GUI_PREFS_CPU_CPU:
			prefs->priv->cpu = g_value_get_int (value);
			Config.Cpu = prefs->priv->cpu;
			break;
		case PROP_GUI_PREFS_CPU_DBG:
			prefs->priv->dbg = g_value_get_int (value);
			Config.Dbg = prefs->priv->dbg;
			break;
		case PROP_GUI_PREFS_CPU_PSXOUT:
			prefs->priv->psxout = g_value_get_int (value);
			Config.PsxOut = prefs->priv->psxout;
			break;
		case PROP_GUI_PREFS_CPU_SPUIRQ:
			prefs->priv->spuirq = g_value_get_int (value);
			Config.SpuIrq = prefs->priv->spuirq;
			break;
		case PROP_GUI_PREFS_CPU_RCNTFIX:
			prefs->priv->rcntfix = g_value_get_int (value);
			Config.RCntFix = prefs->priv->rcntfix;
			break;
		case PROP_GUI_PREFS_CPU_VSYNCWA:
			prefs->priv->vsyncwa = g_value_get_int (value);
			Config.VSyncWA = prefs->priv->vsyncwa;
			break;
		case PROP_GUI_PREFS_MEMCARD1:
			prefs->priv->memcard1 = g_strdup (g_value_get_string (value));
			g_snprintf(Config.Mcd1, MAXPATHLEN, "%s", prefs->priv->memcard1);
			break;
		case PROP_GUI_PREFS_MEMCARD2:
			prefs->priv->memcard2 = g_strdup (g_value_get_string (value));
			g_snprintf(Config.Mcd2, MAXPATHLEN, "%s", prefs->priv->memcard2);
			break;
		case PROP_GUI_PREFS_CURRENTFILE:
			prefs->priv->currentfile = g_strdup (g_value_get_string (value));
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}

	/* As long as the notify::<property> callback is set, the handler
	   will then manage saving the key file */

}

static void
pcsx_gui_prefs_get_property (GObject *object,
			     guint prop_id,
			     GValue *value,
			     GParamSpec *pspec)
{
	PcsxGuiPrefs *prefs = PCSX_GUI_PREFS (object);

	switch (prop_id) {

		case PROP_GUI_PREFS_DIR_BIOS:
			g_value_set_string (value, prefs->priv->biosdir);
			break;
		case PROP_GUI_PREFS_DIR_PLUGINS:
			g_value_set_string (value, prefs->priv->plugindir);
			break;
		case PROP_GUI_PREFS_BIOS:
			g_value_set_string (value, prefs->priv->bios);
			break;
		case PROP_GUI_PREFS_LOADCDBIOS:
			g_value_set_int (value, prefs->priv->loadcdbios);
			break;
		case PROP_GUI_PREFS_PLUGIN_GPU:
			g_value_set_string (value, prefs->priv->gpu_plugin);
			break;
		case PROP_GUI_PREFS_PLUGIN_SPU:
			g_value_set_string (value, prefs->priv->spu_plugin);
			break;
		case PROP_GUI_PREFS_PLUGIN_PAD1:
			g_value_set_string (value, prefs->priv->pad1_plugin);
			break;
		case PROP_GUI_PREFS_PLUGIN_PAD2:
			g_value_set_string (value, prefs->priv->pad2_plugin);
			break;
		case PROP_GUI_PREFS_PLUGIN_CDR:
			g_value_set_string (value, prefs->priv->cdr_plugin);
			break;
		case PROP_GUI_PREFS_CPU_XA:
			g_value_set_int (value, prefs->priv->xa);
			break;
		case PROP_GUI_PREFS_CPU_SIO:
			g_value_set_int (value, prefs->priv->sio);
			break;
		case PROP_GUI_PREFS_CPU_MDEC:
			g_value_set_int (value, prefs->priv->mdec);
			break;
		case PROP_GUI_PREFS_CPU_PSXAUTO:
			g_value_set_int (value, prefs->priv->psxauto);
			break;
		case PROP_GUI_PREFS_CPU_PSXTYPE:
			g_value_set_int (value, prefs->priv->psxtype);
			break;
		case PROP_GUI_PREFS_CPU_CDDA:
			g_value_set_int (value, prefs->priv->cdda);
			break;
		case PROP_GUI_PREFS_CPU_CPU:
			g_value_set_int (value, prefs->priv->cpu);
			break;
		case PROP_GUI_PREFS_CPU_DBG:
			g_value_set_int (value, prefs->priv->dbg);
			break;
		case PROP_GUI_PREFS_CPU_PSXOUT:
			g_value_set_int (value, prefs->priv->psxout);
			break;
		case PROP_GUI_PREFS_CPU_SPUIRQ:
			g_value_set_int (value, prefs->priv->spuirq);
			break;
		case PROP_GUI_PREFS_CPU_RCNTFIX:
			g_value_set_int (value, prefs->priv->rcntfix);
			break;
		case PROP_GUI_PREFS_CPU_VSYNCWA:
			g_value_set_int (value, prefs->priv->vsyncwa);
			break;
		case PROP_GUI_PREFS_MEMCARD1:
			g_value_set_string (value, prefs->priv->memcard1);
			break;
		case PROP_GUI_PREFS_MEMCARD2:
			g_value_set_string (value, prefs->priv->memcard2);
			break;
		case PROP_GUI_PREFS_CURRENTFILE:
			g_value_set_string (value, prefs->priv->currentfile);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static void
pcsx_gui_prefs_finalize (GObject *obj)
{
	printf ("Finalising pcsx_gui_prefs object\n");
	
	PcsxGuiPrefs *pr = PCSX_GUI_PREFS (obj);
	
	/* Free the key file for the preferences ini */
	if (pr->priv->prefs_ini_file)
		g_key_file_free (pr->priv->prefs_ini_file);
	g_free (pr->priv->filename);
	
	g_free (pr->priv->biosdir);
	g_free (pr->priv->plugindir);
	
	g_free (pr->priv->bios);
	g_free (pr->priv->gpu_plugin);
	g_free (pr->priv->spu_plugin);
	g_free (pr->priv->pad1_plugin);
	g_free (pr->priv->pad2_plugin);
	g_free (pr->priv->cdr_plugin);
	
	g_free (pr->priv->memcard1);
	g_free (pr->priv->memcard2);
	
	g_free (pr->priv->currentfile);
	
	g_free (pr->priv);
	
	printf ("Finalising pcsx_gui_prefs object... done\n");

}

static void
pcsx_gui_prefs_class_init (PcsxGuiPrefsClass *klass)
{
	
	
	GObjectClass *object_class = G_OBJECT_CLASS (klass);
	
	object_class->set_property = pcsx_gui_prefs_set_property;
	object_class->get_property = pcsx_gui_prefs_get_property;
	object_class->finalize = pcsx_gui_prefs_finalize;

	/* UI preferences */
	g_object_class_install_property (object_class,
					 PROP_GUI_PREFS_DIR_BIOS,
					 g_param_spec_string ("dir-bios", "", "", PACKAGE_DATA_DIR "bios", G_PARAM_READWRITE));
	g_object_class_install_property (object_class,
					 PROP_GUI_PREFS_DIR_PLUGINS,
					 g_param_spec_string ("dir-plugins", "", "", "/usr/lib/games/psemu/"/*PACKAGE_DATA_DIR "plugins"*/, G_PARAM_READWRITE));
	g_object_class_install_property (object_class,
					 PROP_GUI_PREFS_BIOS,
					 g_param_spec_string ("bios-bios", "", "", "HLE", G_PARAM_READWRITE));
	g_object_class_install_property (object_class,
					 PROP_GUI_PREFS_LOADCDBIOS,
					 g_param_spec_int ("bios-loadcdbios", "", "", 0, 1, 1, G_PARAM_READWRITE));
	g_object_class_install_property (object_class,
					 PROP_GUI_PREFS_PLUGIN_GPU,
					 g_param_spec_string ("plugin-gpu", "", "", NULL, G_PARAM_READWRITE));  // FIXME - default value
	g_object_class_install_property (object_class,
					 PROP_GUI_PREFS_PLUGIN_SPU,
					 g_param_spec_string ("plugin-spu", "", "", NULL, G_PARAM_READWRITE));  // FIXME - default value
	g_object_class_install_property (object_class,
					 PROP_GUI_PREFS_PLUGIN_PAD1,
					 g_param_spec_string ("plugin-pad1", "", "", NULL, G_PARAM_READWRITE));  // FIXME - default value
	g_object_class_install_property (object_class,
					 PROP_GUI_PREFS_PLUGIN_PAD2,
					 g_param_spec_string ("plugin-pad2", "", "", NULL, G_PARAM_READWRITE));  // FIXME - default value
	g_object_class_install_property (object_class,
					 PROP_GUI_PREFS_PLUGIN_CDR,
					 g_param_spec_string ("plugin-cdr", "", "", NULL, G_PARAM_READWRITE));  // FIXME - default value
	g_object_class_install_property (object_class,
					 PROP_GUI_PREFS_CPU_XA,
					 g_param_spec_int ("cpu-xa", "", "", 0, 1, 0, G_PARAM_READWRITE));
	g_object_class_install_property (object_class,
					 PROP_GUI_PREFS_CPU_SIO,
					 g_param_spec_int ("cpu-sio", "", "", 0, 1, 0, G_PARAM_READWRITE));
	g_object_class_install_property (object_class,
					 PROP_GUI_PREFS_CPU_MDEC,
					 g_param_spec_int ("cpu-mdec", "", "", 0, 1, 0, G_PARAM_READWRITE));
	g_object_class_install_property (object_class,
					 PROP_GUI_PREFS_CPU_PSXAUTO,
					 g_param_spec_int ("cpu-psxauto", "", "", 0, 1, 0, G_PARAM_READWRITE));
	g_object_class_install_property (object_class,
					 PROP_GUI_PREFS_CPU_PSXTYPE,
					 g_param_spec_int ("cpu-psxtype", "", "", 0, 1, PSX_TYPE_NTSC, G_PARAM_READWRITE));
	g_object_class_install_property (object_class,
					 PROP_GUI_PREFS_CPU_CDDA,
					 g_param_spec_int ("cpu-cdda", "", "", 0, 1, 0, G_PARAM_READWRITE));
	g_object_class_install_property (object_class,
					 PROP_GUI_PREFS_CPU_CPU,
					 g_param_spec_int ("cpu-cpu", "", "", 0, 1, 0, G_PARAM_READWRITE));
	g_object_class_install_property (object_class,
					 PROP_GUI_PREFS_CPU_DBG,
					 g_param_spec_int ("cpu-dbg", "", "", 0, 1, 0, G_PARAM_READWRITE));
	g_object_class_install_property (object_class,
					 PROP_GUI_PREFS_CPU_PSXOUT,
					 g_param_spec_int ("cpu-psxout", "", "", 0, 1, 0, G_PARAM_READWRITE));
	g_object_class_install_property (object_class,
					 PROP_GUI_PREFS_CPU_SPUIRQ,
					 g_param_spec_int ("cpu-spuirq", "", "", 0, 1, 0, G_PARAM_READWRITE));
	g_object_class_install_property (object_class,
					 PROP_GUI_PREFS_CPU_RCNTFIX,
					 g_param_spec_int ("cpu-rcntfix", "", "", 0, 1, 0, G_PARAM_READWRITE));
	g_object_class_install_property (object_class,
					 PROP_GUI_PREFS_CPU_VSYNCWA,
					 g_param_spec_int ("cpu-vsyncwa", "", "", 0, 1, 0, G_PARAM_READWRITE));
	g_object_class_install_property (object_class,
					 PROP_GUI_PREFS_MEMCARD1,
					 g_param_spec_string ("memcard-memcard1", "", "", g_build_filename (g_get_home_dir (), ".pcsx", "memcards", "card1.mcd", NULL), G_PARAM_READWRITE));
	g_object_class_install_property (object_class,
					 PROP_GUI_PREFS_MEMCARD2,
					 g_param_spec_string ("memcard-memcard2", "", "", g_build_filename (g_get_home_dir (), ".pcsx", "memcards", "card1.mcd", NULL), G_PARAM_READWRITE));
	g_object_class_install_property (object_class,
					 PROP_GUI_PREFS_CURRENTFILE,
					 g_param_spec_string ("file-currentfile", "", "", NULL, G_PARAM_READWRITE));
	
}

static void
pcsx_gui_prefs_init (PcsxGuiPrefs *pr)
{
	gboolean load_result;
	
	printf("pcsx_gui_prefs_init starting...\n");
	
	pr->priv = g_new0 (PcsxGuiPrefsPrivate, 1);
	
	pr->priv->filename = g_build_filename (g_get_home_dir (), ".pcsx", "pcsx.ini", NULL);

	pr->priv->prefs_ini_file = g_key_file_new ();
	GError *error = NULL;
	load_result = g_key_file_load_from_file (pr->priv->prefs_ini_file, pr->priv->filename,
						 G_KEY_FILE_KEEP_COMMENTS, &error);

	if (!load_result) {
		printf ("Error loading %s - %s\n", pr->priv->filename, error->message);	
	
		g_error_free (error);
		error = NULL;
	}
	
	/* Plugin preferences */
	pr->priv->biosdir = pcsx_gui_prefs_get_string_property_from_key_file (pr, "dir-bios");
	pr->priv->plugindir = pcsx_gui_prefs_get_string_property_from_key_file (pr, "dir-plugins");
	
	pr->priv->bios = pcsx_gui_prefs_get_string_property_from_key_file (pr, "bios-bios");
	pr->priv->loadcdbios = pcsx_gui_prefs_get_int_property_from_key_file (pr, "bios-loadcdbios");
	pr->priv->gpu_plugin = pcsx_gui_prefs_get_string_property_from_key_file (pr, "plugin-gpu");
	pr->priv->spu_plugin = pcsx_gui_prefs_get_string_property_from_key_file (pr, "plugin-spu");
	pr->priv->pad1_plugin = pcsx_gui_prefs_get_string_property_from_key_file (pr, "plugin-pad1");
	pr->priv->pad2_plugin = pcsx_gui_prefs_get_string_property_from_key_file (pr, "plugin-pad2");
	pr->priv->cdr_plugin = pcsx_gui_prefs_get_string_property_from_key_file (pr, "plugin-cdr");

	/* UI preferences */
	pr->priv->xa = pcsx_gui_prefs_get_int_property_from_key_file (pr, "cpu-xa");
	pr->priv->sio = pcsx_gui_prefs_get_int_property_from_key_file (pr, "cpu-sio");
	pr->priv->mdec = pcsx_gui_prefs_get_int_property_from_key_file (pr, "cpu-mdec");
	pr->priv->psxauto = pcsx_gui_prefs_get_int_property_from_key_file (pr, "cpu-psxauto");
	pr->priv->psxtype = pcsx_gui_prefs_get_int_property_from_key_file (pr, "cpu-psxtype");
	pr->priv->cdda = pcsx_gui_prefs_get_int_property_from_key_file (pr, "cpu-cdda");
	pr->priv->cpu = pcsx_gui_prefs_get_int_property_from_key_file (pr, "cpu-cpu");
	pr->priv->dbg = pcsx_gui_prefs_get_int_property_from_key_file (pr, "cpu-dbg");
	pr->priv->psxout = pcsx_gui_prefs_get_int_property_from_key_file (pr, "cpu-psxout");
	pr->priv->spuirq = pcsx_gui_prefs_get_int_property_from_key_file (pr, "cpu-spuirq");
	pr->priv->rcntfix = pcsx_gui_prefs_get_int_property_from_key_file (pr, "cpu-rcntfix");
	pr->priv->vsyncwa = pcsx_gui_prefs_get_int_property_from_key_file (pr, "cpu-vsyncwa");
	
	pr->priv->memcard1 = pcsx_gui_prefs_get_string_property_from_key_file (pr, "memcard-memcard1");
	pr->priv->memcard2 = pcsx_gui_prefs_get_string_property_from_key_file (pr, "memcard-memcard2");
	
	/* Set the Config.* values */
	g_snprintf(Config.BiosDir, MAXPATHLEN, "%s", pr->priv->biosdir);
	g_snprintf(Config.PluginsDir, MAXPATHLEN, "%s", pr->priv->plugindir);
	g_snprintf(Config.Bios, 255, "%s", pr->priv->bios);
	g_snprintf(Config.Gpu, 255, "%s", pr->priv->gpu_plugin);
	g_snprintf(Config.Spu, 255, "%s", pr->priv->spu_plugin);
	g_snprintf(Config.Cdr, 255, "%s", pr->priv->cdr_plugin);
	g_snprintf(Config.Pad1, 255, "%s", pr->priv->pad1_plugin);
	g_snprintf(Config.Pad2, 255, "%s", pr->priv->pad2_plugin);
	
	Config.Xa = pr->priv->xa;
	Config.Sio = pr->priv->sio;
	Config.Mdec = pr->priv->mdec;
	Config.PsxAuto = pr->priv->psxauto;
	Config.PsxType = pr->priv->psxtype;
	Config.Cdda = pr->priv->cdda;
	Config.Cpu = pr->priv->cpu;
	Config.Dbg = pr->priv->dbg;
	Config.PsxOut = pr->priv->psxout;
	Config.SpuIrq = pr->priv->spuirq;
	Config.RCntFix = pr->priv->rcntfix;
	Config.VSyncWA = pr->priv->vsyncwa;
	
	g_snprintf(Config.Mcd1, MAXPATHLEN, "%s", pr->priv->memcard1);
	g_snprintf(Config.Mcd2, MAXPATHLEN, "%s", pr->priv->memcard2);

	/* Set handlers so that whenever the values are changed (from anywhere), the signal handler
	   is invoked; the callback then saves to the g_key_file */
	g_signal_connect (pr, "notify::dir-bios",  (GCallback) pcsx_gui_prefs_set_string_property_to_key_file, NULL);
	g_signal_connect (pr, "notify::dir-plugins",  (GCallback) pcsx_gui_prefs_set_string_property_to_key_file, NULL);
	g_signal_connect (pr, "notify::bios-bios",  (GCallback) pcsx_gui_prefs_set_string_property_to_key_file, NULL);
	g_signal_connect (pr, "notify::bios-loadcdbios",  (GCallback) pcsx_gui_prefs_set_int_property_to_key_file, NULL);
	g_signal_connect (pr, "notify::plugin-gpu",  (GCallback) pcsx_gui_prefs_set_string_property_to_key_file, NULL);
	g_signal_connect (pr, "notify::plugin-spu",  (GCallback) pcsx_gui_prefs_set_string_property_to_key_file, NULL);
	g_signal_connect (pr, "notify::plugin-pad1",  (GCallback) pcsx_gui_prefs_set_string_property_to_key_file, NULL);
	g_signal_connect (pr, "notify::plugin-pad2",  (GCallback) pcsx_gui_prefs_set_string_property_to_key_file, NULL);
	g_signal_connect (pr, "notify::plugin-cdr",  (GCallback) pcsx_gui_prefs_set_string_property_to_key_file, NULL);
	g_signal_connect (pr, "notify::cpu-xa",  (GCallback) pcsx_gui_prefs_set_int_property_to_key_file, NULL);
	g_signal_connect (pr, "notify::cpu-sio",  (GCallback) pcsx_gui_prefs_set_int_property_to_key_file, NULL);
	g_signal_connect (pr, "notify::cpu-mdec",  (GCallback) pcsx_gui_prefs_set_int_property_to_key_file, NULL);
	g_signal_connect (pr, "notify::cpu-psxauto",  (GCallback) pcsx_gui_prefs_set_int_property_to_key_file, NULL);
	g_signal_connect (pr, "notify::cpu-psxtype",  (GCallback) pcsx_gui_prefs_set_int_property_to_key_file, NULL);
	g_signal_connect (pr, "notify::cpu-cdda",  (GCallback) pcsx_gui_prefs_set_int_property_to_key_file, NULL);
	g_signal_connect (pr, "notify::cpu-cpu",  (GCallback) pcsx_gui_prefs_set_int_property_to_key_file, NULL);
	g_signal_connect (pr, "notify::cpu-dbg",  (GCallback) pcsx_gui_prefs_set_int_property_to_key_file, NULL);
	g_signal_connect (pr, "notify::cpu-psxout",  (GCallback) pcsx_gui_prefs_set_int_property_to_key_file, NULL);
	g_signal_connect (pr, "notify::cpu-spuirq",  (GCallback) pcsx_gui_prefs_set_int_property_to_key_file, NULL);
	g_signal_connect (pr, "notify::cpu-rcntfix", (GCallback) pcsx_gui_prefs_set_int_property_to_key_file, NULL);
	g_signal_connect (pr, "notify::cpu-vsyncwa", (GCallback) pcsx_gui_prefs_set_int_property_to_key_file, NULL);
	g_signal_connect (pr, "notify::memcard-memcard1",  (GCallback) pcsx_gui_prefs_set_string_property_to_key_file, NULL);
	g_signal_connect (pr, "notify::memcard-memcard2",  (GCallback) pcsx_gui_prefs_set_string_property_to_key_file, NULL);

	printf("pcsx_gui_prefs_init finished\n");
}

PcsxGuiPrefs* pcsx_gui_prefs_new (void)
{
	return g_object_new (PCSX_TYPE_GUI_PREFS, NULL);
}

static void
pcsx_gui_prefs_set_int_property_to_key_file (PcsxGuiPrefs *pr, GParamSpec *param, gpointer user_data)
{
	gchar *key;
	gchar **category_key_pair;
	gint value;
	GError *error = NULL;

	g_return_if_fail (pr->priv->prefs_ini_file != NULL);
	g_return_if_fail (pr->priv->filename != NULL);
	
	/* Get the key, which will be in the format category.key */
	key = g_strdup (g_param_spec_get_name (param));
	
	g_return_if_fail (key != NULL);	
	
	g_object_get (pr, key, &value, NULL);
	
	/*printf ("Saving %s - setting integer value %s with value %d\n",
		       pr->priv->filename, key, value);*/

	/* Set the value and save the file */
	category_key_pair = g_strsplit (key, "-", 2);
	g_key_file_set_integer (pr->priv->prefs_ini_file, category_key_pair[0], category_key_pair[1], value);
	g_key_file_save_to_file (pr->priv->prefs_ini_file, pr->priv->filename, &error);
	
	if (error) {
		printf ("Error saving %s - %s\n", pr->priv->filename, error->message);
		g_error_free (error);
	}
	
	g_free (key);
	g_strfreev (category_key_pair);
}

static void
pcsx_gui_prefs_set_string_property_to_key_file (PcsxGuiPrefs *pr, GParamSpec *param, gpointer user_data)
{
	gchar *key;
	gchar **category_key_pair;
	gchar *value;
	GError *error = NULL;

	g_return_if_fail (pr->priv->prefs_ini_file != NULL);
	g_return_if_fail (pr->priv->filename != NULL);
	
	/* Get the key, which will be in the format category.key */
	key = g_strdup (g_param_spec_get_name (param));
	
	g_return_if_fail (key != NULL);	
	
	g_object_get (pr, key, &value, NULL);
	
	/*printf ("Saving %s - setting string value %s with value %s\n",
		       pr->priv->filename, key, value);*/
	g_return_if_fail (value != NULL);

	/* Set the value and save the file */
	category_key_pair = g_strsplit (key, "-", 2);
	g_key_file_set_string (pr->priv->prefs_ini_file, category_key_pair[0], category_key_pair[1], value);
	g_key_file_save_to_file (pr->priv->prefs_ini_file, pr->priv->filename, &error);
	
	if (error) {
		printf ("Error saving %s - %s\n", pr->priv->filename, error->message);
		g_error_free (error);
	}
	
	g_free (key);
	g_strfreev (category_key_pair);
}

/* Retrieve an integer property from the preferences ini file. If an error occurs (usually because
   the preferences file does not exist, or the key value is not present, then the default value
   in the g_param_spec is used instead */
static gint
pcsx_gui_prefs_get_int_property_from_key_file (PcsxGuiPrefs *pr, gchar *property) {
	GError *error = NULL;
	gint val;
	
	gchar **category_key_pair;
	category_key_pair = g_strsplit (property, "-", 2);
	
	val = g_key_file_get_integer (pr->priv->prefs_ini_file, category_key_pair[0], category_key_pair[1], &error);

	if (error) {
		printf ("Error retrieving integer UI option %s - %s\n", property, error->message);
		g_error_free (error);
		error = NULL;

		GParamSpec *spec;
		GValue value = { 0, };

		spec = g_object_class_find_property (G_OBJECT_GET_CLASS (pr), property);
		
		g_value_init (&value, G_TYPE_INT);
		g_param_value_set_default (spec, &value);
		
		val = g_value_get_int (&value);
		/*printf (_("Retrieving default integer value for %s: %d\n"), property, val);*/
	}
	
	g_strfreev (category_key_pair);
	
	return val;
}

static gchar*
pcsx_gui_prefs_get_string_property_from_key_file (PcsxGuiPrefs *pr, gchar *property) {
	GError *error = NULL;
	gchar* val;
	
	gchar **category_key_pair;
	category_key_pair = g_strsplit (property, "-", 2);
	
	val = g_key_file_get_string (pr->priv->prefs_ini_file, category_key_pair[0], category_key_pair[1], &error);

	if (error) {
		printf ("Error retrieving string UI option %s - %s\n", property, error->message);
		g_error_free (error);
		error = NULL;

		GParamSpec *spec;
		GValue value = { 0, };

		spec = g_object_class_find_property (G_OBJECT_GET_CLASS (pr), property);
		
		g_value_init (&value, G_TYPE_STRING);
		g_param_value_set_default (spec, &value);
		
		val = g_value_get_string (&value);
		/*printf (_("Retrieving default string value for %s: %s\n"), property, val);*/
	}
	
	g_strfreev (category_key_pair);
	
	return val;
}

/* Custom function to save the keyfile to disk. If the file does not exist,
   it will be created */
gboolean
g_key_file_save_to_file(GKeyFile *kf, const gchar *file, GError **error)
{
	gsize length, nr;
	char *data = g_key_file_to_data(kf, &length, error);
	GIOChannel *gio;

	if (!data)
		return FALSE;

	gio = g_io_channel_new_file(file, "w+", error);
	if (!gio) {
		g_free(data);
		return FALSE;
	}
	
	g_io_channel_write_chars(gio, data, length, &nr, error);

	g_free(data);

	g_io_channel_unref(gio);

	return TRUE;
}




#ifdef REDUNDANT_CODE
* TODO escaping/unescaping would be nice, as would maxchars *
void GetValue(char *src, char *name, char *outvar)
{
	char *tmp;

	*outvar = 0;
	tmp = strstr(src, name); 
	if (tmp == NULL) return;

	tmp+=strlen(name); 
	while ((*tmp == ' ') || (*tmp == '=')) tmp++; 

	while (*tmp != '\n' && *tmp != 0)
		*outvar++ = *tmp++;

	*outvar = 0;
	return;
}

void GetValuel(char *src, char *name, long *var) 
{
	char *tmp = strstr(src, name); 
	if (tmp != NULL) { 
		tmp+=strlen(name); 
		while ((*tmp == ' ') || (*tmp == '=')) tmp++; 
		if (*tmp != '\n') sscanf(tmp, "%lx", var); 
	}
	return;
}

#define SetValue(name, var) \
	fprintf (f,"%s = %s\n", name, var);

#define SetValuel(name, var) \
	fprintf (f,"%s = %lx\n", name, var);

int LoadConfig(PcsxConfig *Conf) {
	struct stat buf;
	FILE *f;
	int size;
	char *data,*tmp;

	/* TODO local var called cfgfile */
	
    // Ryan says: use dotdir, dotdir is GOOD
    // No giant homedir names
    strncpy(cfgfile, getenv("HOME"), 200);
    strcat(cfgfile, PCSX_DOT_DIR);
    
	// FIXME: this should check for each subdirectory's existence as well
	//        otherwise upgraded installs won't have a working dotdir!
    // make sure that ~/.pcsx exists
    if (stat(cfgfile, &buf) == -1) {
        // if it doesn't, make it and the subdirs
        // this is so ugly, I miss Python

        mkdir(cfgfile, S_IRWXU|S_IRWXG);
    }
    check_sub_directories ();

    // proceed to load the cfg file
    // append its name
    strcat(cfgfile, cfgfile_basename);

    // file is  now ~/.pcsx/pcsx.cfg (or whatever cfgfile_basename is)
    if (stat(cfgfile, &buf) == -1) {
        // the config file doesn't exist!
		/* TODO Error checking? */
		printf("Configuration file %s couldn't be found\n", cfgfile);
        return -1;
	}
	
	size = buf.st_size;

	/* TODO Error checking for the next two lines, and at least log failures */
	f = fopen(cfgfile, READONLY);
	if (f == NULL) return -1;

	data = (char*)malloc(size);
	if (data == NULL) return -1;

	fread(data, 1, buf.st_size, f);
	fclose(f);

	GetValue(data, "Bios", Config.Bios);
	GetValue(data, "Gpu",  Config.Gpu);
	GetValue(data, "Spu",  Config.Spu);
	GetValue(data, "Cdr",  Config.Cdr);
	GetValue(data, "Pad1", Config.Pad1);
	GetValue(data, "Pad2", Config.Pad2);
	GetValue(data, "Net",  Config.Net);
	GetValue(data, "Mcd1", Config.Mcd1);
	GetValue(data, "Mcd2", Config.Mcd2);
	GetValue(data, "BiosDir",    Config.BiosDir);
	GetValue(data, "PluginsDir",    Config.PluginsDir);
	GetValuel(data, "Xa",      &Config.Xa);
	GetValuel(data, "Sio",     &Config.Sio);
	GetValuel(data, "Mdec",    &Config.Mdec);
	GetValuel(data, "PsxAuto", &Config.PsxAuto);
	GetValuel(data, "PsxType", &Config.PsxType);
	GetValuel(data, "Cdda",    &Config.Cdda);
	GetValuel(data, "Cpu",     &Config.Cpu);
	GetValuel(data, "Dbg",     &Config.Dbg);
	GetValuel(data, "PsxOut",  &Config.PsxOut);
	GetValuel(data, "SpuIrq",  &Config.SpuIrq);
	GetValuel(data, "RCntFix", &Config.RCntFix);
	GetValuel(data, "VSyncWA", &Config.VSyncWA);

	free(data);

	return 0;
}

/////////////////////////////////////////////////////////

void SaveConfig() {
	FILE *f;

	/* TODO Error checking for the next two lines, and at least log failures - suggest a file dialog to specify a new file or create a new file */
	f = fopen(cfgfile, READWRITE);
	if (f == NULL) return;

	SetValue("Bios", Config.Bios);
	SetValue("Gpu",  Config.Gpu);
	SetValue("Spu",  Config.Spu);
	SetValue("Cdr",  Config.Cdr);
	SetValue("Net",  Config.Net);
	SetValue("Pad1", Config.Pad1);
	SetValue("Pad2", Config.Pad2);
	SetValue("Mcd1", Config.Mcd1);
	SetValue("Mcd2", Config.Mcd2);
	SetValue("BiosDir",    Config.BiosDir);
	SetValue("PluginsDir",    Config.PluginsDir);
	SetValuel("Xa",      Config.Xa);
	SetValuel("Sio",     Config.Sio);
	SetValuel("Mdec",    Config.Mdec);
	SetValuel("PsxAuto", Config.PsxAuto);
	SetValuel("PsxType", Config.PsxType);
	SetValuel("Cdda",    Config.Cdda);
	SetValuel("Cpu",     Config.Cpu);
	SetValuel("Dbg",     Config.Dbg);
	SetValuel("PsxOut",  Config.PsxOut);
	SetValuel("SpuIrq",  Config.SpuIrq);
	SetValuel("RCntFix", Config.RCntFix);
	SetValuel("VSyncWA", Config.VSyncWA);

	fclose(f);
}
#endif /* REDUNDANT_CODE */