/*  Pcsx - Pc Psx Emulator
 *  Copyright (C) 1999-2002  Pcsx Team
 *  Copyright (C) 2008 Andrew Burton
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <glade/glade.h>
#include "Config.h"
#include "configure-dlg.h"

struct _PcsxProperty
{
	GtkWidget    *object;		/* Widget */
	gchar        *key;			/* The string containing the "category-key" */
	gint        *default_value;
	PcsxPropertyObjectType  object_type;	/* Type of the object */
	
};
typedef struct _PcsxProperty PcsxProperty;

struct _PcsxConfigureDialogPrivate {
	GladeXML *xml;

	/* Internal widget references */
	GtkWidget *system_type_combo;	/* NTSC or PAL */
	GtkWidget *bios_combo;

};

#define PCSX_CONFIGURE_DIALOG_GET_PRIVATE(o)  (PCSX_CONFIGURE_DIALOG (o)->priv)

G_DEFINE_TYPE (PcsxConfigureDialog, pcsx_configure_dialog, GTK_TYPE_DIALOG)

/* Function prototypes */
static void
pcsx_configure_dialog_response             (GtkDialog *dialog, gint response);
static void
pcsx_configure_dialog_destroy              (GtkObject *object);

static void
connect_prop_to_object (PcsxProperty *p);
static PcsxProperty*
parse_pcsx_property_from_widget_name (GtkWidget *widget, gchar *widget_name);

static PcsxPropertyObjectType
get_object_type_from_string (const gchar* object_type)
{
	if (strcmp (object_type, "combo") == 0)
		return PCSX_PROPERTY_OBJECT_TYPE_COMBO;
	else if (strcmp (object_type, "bioscombo") == 0)
		return PCSX_PROPERTY_OBJECT_TYPE_BIOSCOMBO;
	else if (strcmp (object_type, "toggle") == 0)
		return PCSX_PROPERTY_OBJECT_TYPE_TOGGLE;
	else if (strcmp (object_type, "confbutton") == 0)
		return PCSX_PROPERTY_OBJECT_TYPE_CONF_BUTTON;
	else if (strcmp (object_type, "aboutbutton") == 0)
		return PCSX_PROPERTY_OBJECT_TYPE_ABOUT_BUTTON;
	else if (strcmp (object_type, "folder") == 0)
		return PCSX_PROPERTY_OBJECT_TYPE_FOLDER;

	else
		return (PcsxPropertyObjectType)(-1);
}

/* Sets the widget with the value. Both the widget and value are contained
   in the property passed as a parameter */
static void
set_widget_value_as_string (PcsxProperty *prop, const gchar *value)
{
	gint  int_value;
	gdouble double_value;
	char** values;
	gint i; 
	
	switch (prop->object_type)
	{
		case PCSX_PROPERTY_OBJECT_TYPE_TOGGLE:
			if (value) 
				int_value = atoi (value);
			else
				int_value = 0;

			gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (prop->object),
				                      int_value);
			break;
		case PCSX_PROPERTY_OBJECT_TYPE_COMBO:
		case PCSX_PROPERTY_OBJECT_TYPE_BIOSCOMBO:
			values = g_object_get_data(G_OBJECT(prop->object), "untranslated");
			g_return_if_fail (values != NULL);
			g_return_if_fail (value != NULL);

			for (i=0; values[i] != NULL; i++)
			{
				if (strcmp(value, values[i]) == 0)
				{
					gtk_combo_box_set_active(GTK_COMBO_BOX(prop->object), i);
					break;
				}
			}

			break;
		case PCSX_PROPERTY_OBJECT_TYPE_FOLDER:
			g_return_if_fail (value != NULL);
			
			gtk_file_chooser_set_current_folder (GTK_FILE_CHOOSER (prop->object), value);
			break;
		default:
			g_assert_not_reached ();
			break;

		}
}

/* Gets the value of the widget within the PcsxProperty passed in */
static gchar*
get_widget_value_as_string (PcsxProperty *prop)
{
	GtkTreeIter iter;
	GtkTreeModel *model;
	
	/* Variables for the possible values */
	gint  int_value;
	gdouble double_value;
	gchar** values;
	gchar *text_value = NULL;
	gint idx;

	switch (prop->object_type)
	{
		case PCSX_PROPERTY_OBJECT_TYPE_TOGGLE:
			int_value = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (prop->object));
			text_value = g_strdup_printf ("%d", int_value);
			break;
		case PCSX_PROPERTY_OBJECT_TYPE_COMBO:
		case PCSX_PROPERTY_OBJECT_TYPE_BIOSCOMBO:
			g_return_val_if_fail ((gtk_combo_box_get_active_iter(GTK_COMBO_BOX(prop->object), &iter) != FALSE), NULL);
			
			/* Get the value for the .so at position 1 from the model */
			model = gtk_combo_box_get_model(GTK_COMBO_BOX(prop->object));
			gtk_tree_model_get(model, &iter, 1, &text_value, -1);

			break;
		case PCSX_PROPERTY_OBJECT_TYPE_FOLDER:
			text_value = gtk_file_chooser_get_current_folder (GTK_FILE_CHOOSER (prop->object));
			break;
		default:
			g_assert_not_reached ();
			break;
	}
	
	if (text_value && (strlen (text_value) == 0))
	{
		g_free (text_value);
		text_value = NULL;
	}
	return text_value;
}

/* This function calls the function that gets the value from the widget,
   and converts it to an integer */
static gint
get_widget_value_as_int (PcsxProperty *prop)
{
	gint  int_value;
	gchar *text_value;
	text_value = get_widget_value_as_string (prop);
	int_value = atoi (text_value);
	g_free (text_value);
	
	return int_value;
}

/* Callback when widgets holding integer properties are changed */
static void
update_property_on_change_int (GtkWidget *widget, gpointer user_data)
{
	PcsxProperty *p;
	gint val;
	
	p = (PcsxProperty *) user_data;
	
	/* Get the value of the widget */
	val = get_widget_value_as_int (p);
	
	/* Update the object so that the change is stored */
	/*printf("Updating pcsx_gui_prefs %s with value %d\n", p->key, val);*/
	g_object_set (pcsx_gui_prefs, p->key, val, NULL);

}

/* Callback when widgets holding string properties are changed */
static void
update_property_on_change_str (GtkWidget *widget, gpointer user_data)
{
	PcsxProperty *p;
	gchar *val;

	p = (PcsxProperty *) user_data;
	
	val = get_widget_value_as_string (p);
	
	/* Update the object so that the change is stored */
	/*printf("Updating pcsx_gui_prefs %s with value %s\n", p->key, val);*/
	g_object_set (pcsx_gui_prefs, p->key, val, NULL);
	
	g_free (val);
}

/* Callback when autodetect system type toggle is changed */
static void
on_autodetect_system_toggled (GtkWidget *widget, gpointer user_data)
{
	PcsxConfigureDialog *dialog;
	gboolean toggled;
	
	dialog = (PcsxConfigureDialog *) user_data;
	
	toggled = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget));
	
	/* If system type is set to autodetect, disable the PAL/NTSC combo */
	gtk_widget_set_sensitive (dialog->priv->system_type_combo, !toggled);
}

static void
on_bios_path_changed (GtkWidget *widget, gpointer user_data)
{
	PcsxConfigureDialog *dialog;
	gchar *current_folder;
	gchar *folder;
	PcsxProperty *p;
	
	dialog = (PcsxConfigureDialog *) user_data;
	
	folder = gtk_file_chooser_get_current_folder (GTK_FILE_CHOOSER (widget));
	
	// Check first to see if the current folder is the same as the new one, and
	// abort if so.
	// This works around the problem of the FileChoosers not populating until
	// after everything else, triggering the callback and overriding the prefs
	g_object_get (pcsx_gui_prefs, "dir-bios", &current_folder, NULL);
	if (g_ascii_strcasecmp (current_folder, folder) == 0) {
		g_free (current_folder);
		g_free (folder);
		return;
	}
	
	g_object_set (pcsx_gui_prefs, "dir-bios", folder, NULL);
	
	/* Rescan the BIOS directory */
	scan_bios_dir ();

	const gchar *property = "bioscombo:BIOS:bios-bios";
			
	p = parse_pcsx_property_from_widget_name (dialog->priv->bios_combo, property);
	connect_prop_to_object (p);

	/* Set the default BIOS to HLE, which is always available */
	gtk_combo_box_set_active (GTK_COMBO_BOX (dialog->priv->bios_combo), 0);
	
	g_free (current_folder);
	g_free (folder);
}

static void
on_plugins_path_changed (GtkWidget *widget, gpointer user_data)
{
	PcsxConfigureDialog *dialog;
	gchar *folder;
	gchar *current_folder;

	GList *widget_list;	/* GList of CPU checkbutton and plugin-related widgets */
	GList *node;
	PcsxProperty *p;
	
	dialog = (PcsxConfigureDialog *) user_data;
	
	g_return_if_fail (dialog != NULL);
	
	folder = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (widget));	
	
	/* Check first to see if the current folder is the same as the new one, and
	   abort if so.
	   This works around the problem of the FileChoosers not populating until
	   after everything else, triggering the callback and overriding the prefs */
	g_object_get (pcsx_gui_prefs, "dir-plugins", &current_folder, NULL);

	if (g_ascii_strcasecmp (current_folder, folder) == 0) {
		g_free (current_folder);
		g_free (folder);
		return;
	}
	
	g_object_set (pcsx_gui_prefs, "dir-plugins", folder, NULL);
	
	scan_plugins_dir ();
	plugin_file_handler_symlink_plugins (folder);   /* Check that this works */

	/* Get the combos for each of the plugins */
	widget_list = glade_xml_get_widget_prefix (dialog->priv->xml, "preferences_combo");

	node = widget_list;
	while (node) {
		GtkWidget *widget;
		gchar *name;
		
		widget = node->data;
		name = g_strdup (glade_get_widget_name (node->data));
		
		if (strncmp (name, "preferences_",
                     strlen ("preferences_")) == 0)
		{
			const gchar *property = &name[strlen ("preferences_")];
			
			p = parse_pcsx_property_from_widget_name (widget, property);
			connect_prop_to_object (p);
		}
		
		/* Set default to first element */
		gtk_combo_box_set_active (GTK_COMBO_BOX (widget), 0);
		
		node = g_list_next (node);
	}
	g_list_free (widget_list);
	g_list_free (node);

	g_free (current_folder);
	g_free (folder);
}

static void
on_system_type_combo_changed (GtkWidget *widget, gpointer user_data)
{
	gint system_type;
	
	/* Note: Assumes NTSC is listed first */
	system_type = gtk_combo_box_get_active (GTK_COMBO_BOX (widget));
	g_object_set (pcsx_gui_prefs, "cpu-psxtype", system_type, NULL);

}

int
get_plugin_type (gchar *type)
{
	if (strcmp (type, "GPU") == 0)
		return PSE_LT_GPU;
	else if (strcmp (type, "SPU") == 0)
		return PSE_LT_SPU;
	else if (strcmp (type, "CDR") == 0)
		return PSE_LT_CDR;
	else if (g_ascii_strncasecmp (type, "PAD", 3) == 0) /* Limited to 3 since we need to handle PAD1 and PAD2 */
		return PSE_LT_PAD;
	else {
		/*printf("get_plugin_type: unsupported type %s\n", type);*/
		return 0;
	}
}

static void
on_plugin_conf_clicked (GtkWidget *widget, gpointer user_data)
{
	long (*conf)(); /* Reference to the configure function in the shared library; note
					   the configure function typedefs in libpcsxcore/plugins.h, e.g.
					   typedef long (CALLBACK* PADconfigure)(void); */

	void *drv;
	gchar *plugin;
	gchar *plugin_dir;
	gchar *plugin_param;
	gchar *plugin_filename;
	gchar *conf_call;
	const gchar *plugin_type;

	PcsxProperty *prop = (PcsxProperty *) user_data;
	g_return_if_fail (prop != NULL);

	plugin_type = &prop->key[strlen ("configure-")];
	
	plugin_param = g_strdup_printf ("plugin-%s", plugin_type);  /* Generates e.g. plugin-gpu */

	g_object_get (pcsx_gui_prefs,
				  "dir-plugins", &plugin_dir,
				  plugin_param, &plugin,
				  NULL);

	/* Build the filename (assumes plugin is symlinked to, and executed from,
	   the writeable $HOME dotdir */
	plugin_filename = g_build_filename (g_get_home_dir (), PLUGINS_DIR, plugin, NULL);

	/* Get the config name from the plugin */
	printf("DEBUG: Loading configuration for %s\n", plugin_filename);
	drv = SysLoadLibrary (plugin_filename);
	g_return_if_fail (drv != NULL);
	// printf("Loaded configure from %s\n", plugin_filename);

	/* Strip the 1 or 2 from the end of PAD */
	if (strstr (plugin_type, "pad") != NULL)
		plugin_type = "pad";

	conf_call = g_strdup_printf("%sconfigure", g_ascii_strup (plugin_type, -1));   /* Generates e.g. GPUconfigure */

	conf = SysLoadSym(drv, conf_call);

	if (conf)
		conf();
	else
		SysInfoMessage (_("No configuration required"), _("This plugin doesn't need to be configured."));
	
	SysCloseLibrary (drv);
	
	g_free (plugin);
	g_free (plugin_dir);
	g_free (plugin_param);
	g_free (plugin_filename);
	g_free (conf_call);

}

static void
on_plugin_about_clicked (GtkWidget *widget, gpointer user_data)
{
	void (*about)(); /* Reference to the about function in the shared library; note
					   the about function typedefs in libpcsxcore/plugins.h, e.g.
					   typedef void (CALLBACK* PADabout)(void); */

	void *drv;
	gchar *plugin;
	gchar *plugin_dir;
	gchar *plugin_param;
	gchar *plugin_filename;
	gchar *about_call;
	const gchar *plugin_type;

	PcsxProperty *prop = (PcsxProperty *) user_data;
	g_return_if_fail (prop != NULL);
	
	plugin_type = &prop->key[strlen ("configure-")];
	
	plugin_param = g_strdup_printf ("plugin-%s", plugin_type);  /* Generates e.g. plugin-gpu */
	
	g_object_get (pcsx_gui_prefs,
				  "dir-plugins", &plugin_dir,
				  plugin_param, &plugin,
				  NULL);
	
	plugin_filename = g_build_filename (plugin_dir, plugin, NULL);

	/* Get the config name from the plugin */
	// printf("Loading configure from %s\n", plugin_filename);
	drv = SysLoadLibrary (plugin_filename);
	g_return_if_fail (drv != NULL);
	// printf("Loaded configure from %s\n", plugin_filename);

	/* Strip the 1 or 2 from the end of PAD */
	if (strstr (plugin_type, "pad") != NULL)
		plugin_type = "pad";
	
	about_call = g_strdup_printf("%sabout", g_ascii_strup (plugin_type, -1));   /* Generates e.g. GPUabout */
	about = SysLoadSym(drv, about_call);

	if (about)
		about();
	else
		SysInfoMessage (_("No About information"), _("This plugin does not provide any About information."));
	
	SysCloseLibrary (drv);
	
	g_free (plugin);
	g_free (plugin_dir);
	g_free (plugin_param);
	g_free (plugin_filename);
	g_free (about_call);

}

/* Get the value of the property from the gui_prefs object and set the widget */
static void
connect_prop_to_object (PcsxProperty *p)
{
	int int_value;
	gchar *str_value;
	
	if (p->object_type == PCSX_PROPERTY_OBJECT_TYPE_TOGGLE)
	{	
		g_object_get (pcsx_gui_prefs, p->key, &int_value, NULL);
		str_value = g_strdup_printf ("%d", int_value);
		set_widget_value_as_string (p, str_value);

		g_free (str_value);
	} else if (p->object_type == PCSX_PROPERTY_OBJECT_TYPE_COMBO) {
		GtkListStore *store;
		GtkTreeIter iter;
		GtkCellRenderer *renderer;
		GList *node;
		pcsx_plugin *plugin;
		gchar *str;
		gchar **vstr;
		
		/* The GtkListStore has two items: 0 = display name, 1 = .so name.
		   We present the first in the combo, and store the second */
		store = gtk_list_store_new (2, G_TYPE_STRING, G_TYPE_STRING);
		
		/* Create a renderer for the combo, clearing any existing ones */
		gtk_cell_layout_clear (GTK_CELL_LAYOUT (p->object));
		renderer = gtk_cell_renderer_text_new();
		gtk_cell_layout_pack_start (GTK_CELL_LAYOUT (p->object), renderer, FALSE);
		gtk_cell_layout_add_attribute (GTK_CELL_LAYOUT (p->object), renderer, "text", 0);
		
		str = (gchar *) g_new0 (gchar*, 1);
		
		for (node = g_list_first (plugin_list); node; node = node->next) {
			plugin = (pcsx_plugin *) node->data;
			
			/* Check that the plugin type matches this combo */
			if (plugin->type == p->default_value) {
				/* Append the plugin name to the combo */
				gtk_list_store_append (GTK_LIST_STORE (store), &iter);
				gtk_list_store_set (GTK_LIST_STORE (store), &iter,
									0, plugin->name,
									1, plugin->so_name,
									-1);
				
				str = g_strconcat (str, plugin->so_name, ",", NULL);
			}
		}
		
		/* Link the model to the combo */
		gtk_combo_box_set_model (GTK_COMBO_BOX (p->object), GTK_TREE_MODEL (store));
		
		vstr = g_strsplit (str, ",", 100);
		g_object_set_data(G_OBJECT(p->object), "untranslated", vstr);

		/* Now need to set the current value based on the preferences */
		g_object_get (pcsx_gui_prefs, p->key, &str_value, NULL);		
		set_widget_value_as_string (p, str_value);
		
		g_free (str_value);
		
	} else if (p->object_type == PCSX_PROPERTY_OBJECT_TYPE_BIOSCOMBO) {
		GtkListStore *store;
		GtkTreeIter iter;
		GtkCellRenderer *renderer;
		GList *node;
		pcsx_bios *bios;
		gchar *str;
		gchar **vstr;
		
		/* Create a model for the combo */
		store = gtk_list_store_new (2, G_TYPE_STRING, G_TYPE_STRING);
		
		/* Create a renderer for the combo, clearing any existing ones */
		gtk_cell_layout_clear (GTK_CELL_LAYOUT (p->object));
		renderer = gtk_cell_renderer_text_new();
		gtk_cell_layout_pack_start (GTK_CELL_LAYOUT (p->object), renderer, FALSE);
		gtk_cell_layout_add_attribute (GTK_CELL_LAYOUT (p->object), renderer, "text", 0);
	
		str = (gchar *) g_new0 (gchar*, 1);
		
		for (node = g_list_first (bios_list); node; node = node->next) {

			bios = (pcsx_bios *) node->data;

			/* Append the bios name to the model */
			gtk_list_store_append (GTK_LIST_STORE (store), &iter);
			gtk_list_store_set (GTK_LIST_STORE (store), &iter,
								0, bios->display_name,
								1, bios->internal_name,
								-1);
			
			str = g_strconcat (str, bios->internal_name, ",", NULL);
		}

		/* Link the model to the combo */
		gtk_combo_box_set_model (GTK_COMBO_BOX (p->object), GTK_TREE_MODEL (store));
		
		vstr = g_strsplit (str, ",", 100);
		g_object_set_data(G_OBJECT(p->object), "untranslated", vstr);
		
		/* Now need to set the current value based on the preferences */
		g_object_get (pcsx_gui_prefs, p->key, &str_value, NULL);		
		set_widget_value_as_string (p, str_value);
		
		g_free (str_value);
		
	} else {
		g_object_get (pcsx_gui_prefs, p->key, &str_value, NULL);
		set_widget_value_as_string (p, str_value);

		g_free (str_value);
	}
}

static PcsxProperty*
parse_pcsx_property_from_widget_name (GtkWidget *widget, gchar *widget_name)
{
	PcsxProperty *p;
	gchar **fields;
	
	fields = g_strsplit (widget_name, ":", 3);
	
	gchar *type = fields[0];
	gchar *default_val = fields[1]; /* Needed only for combo and buttons */
	gchar *key = fields[2];
	
	/* Set up the property object so that we can pass it to the callback */
	p = g_new0 (PcsxProperty, 1);
	p->object = widget;
	p->object_type = get_object_type_from_string (type);
	p->default_value = get_plugin_type (default_val);
	p->key = g_strdup (key);
	
	g_strfreev (fields);
	
	return p;
}

static void
register_property_from_glade_string (PcsxConfigureDialog *dialog, GtkWidget *widget, gchar *widget_name)
{
	PcsxProperty *p;

	p = parse_pcsx_property_from_widget_name (widget, widget_name);
	printf ("Processing widget %s\n", widget_name);

	/* Get the key's value and set the widget */
	if ((p->object_type != PCSX_PROPERTY_OBJECT_TYPE_CONF_BUTTON) &&
		(p->object_type != PCSX_PROPERTY_OBJECT_TYPE_ABOUT_BUTTON))
		connect_prop_to_object (p);

	/* Set the callback for the widget */
	switch (p->object_type) {
		case PCSX_PROPERTY_OBJECT_TYPE_TOGGLE:
			g_signal_connect (G_OBJECT(p->object), "toggled",
							  G_CALLBACK (update_property_on_change_int), p);
			break;
		case PCSX_PROPERTY_OBJECT_TYPE_COMBO:
			g_signal_connect (G_OBJECT(p->object), "changed",
							  G_CALLBACK (update_property_on_change_str), p);
			break;
		case PCSX_PROPERTY_OBJECT_TYPE_BIOSCOMBO:
			g_signal_connect (G_OBJECT(p->object), "changed",
							  G_CALLBACK (update_property_on_change_str), p);
			break;
		case PCSX_PROPERTY_OBJECT_TYPE_FOLDER:		
			/* These callbacks also trigger re-population of the associated combos */
			if (strstr (p->key, "dir-bios") != NULL)
				g_signal_connect (G_OBJECT (p->object), "selection-changed",
					  G_CALLBACK (on_bios_path_changed), dialog);
			else {
				g_signal_connect (G_OBJECT (p->object), "selection-changed",
					  G_CALLBACK (on_plugins_path_changed), dialog);
			}
			
			break;
		case PCSX_PROPERTY_OBJECT_TYPE_CONF_BUTTON:
			g_signal_connect (G_OBJECT(p->object), "clicked",
							  G_CALLBACK (on_plugin_conf_clicked), p);
			break;
		case PCSX_PROPERTY_OBJECT_TYPE_ABOUT_BUTTON:
			g_signal_connect (G_OBJECT(p->object), "clicked",
							  G_CALLBACK (on_plugin_about_clicked), p);
			break;
		default:
			printf ("Unsupported object type: %s\n", widget_name);
			break;
	}

}

/* Boilerplate functions */
static GObject *
pcsx_configure_dialog_constructor (GType                  type,
				guint                  n_construct_properties,
				GObjectConstructParam *construct_properties)
{
	GObject          *obj;
	PcsxConfigureDialog *dialog;

	obj = G_OBJECT_CLASS (pcsx_configure_dialog_parent_class)->constructor (type,
									     n_construct_properties,
									     construct_properties);

	dialog = PCSX_CONFIGURE_DIALOG (obj);

	return obj;
}

static void
pcsx_configure_dialog_class_init (PcsxConfigureDialogClass *class)
{
	GObjectClass *gobject_class = G_OBJECT_CLASS (class);
	GtkObjectClass *gtkobject_class = GTK_OBJECT_CLASS (class);
	GtkDialogClass *gtkdialog_class = GTK_DIALOG_CLASS (class);

	gobject_class->constructor = pcsx_configure_dialog_constructor;
/*	gobject_class->get_property = pcsx_configure_dialog_get_property;
	gobject_class->set_property = pcsx_configure_dialog_set_property;*/

	gtkobject_class->destroy = pcsx_configure_dialog_destroy;
	gtkdialog_class->response = pcsx_configure_dialog_response;


	g_type_class_add_private (class,
				  sizeof (PcsxConfigureDialogPrivate));

	/* Signals and properties go here */

}

static void
register_preferences_widgets_from_glade (PcsxConfigureDialog *dialog)
{
	GList *widget_list;	/* GList of CPU checkbutton and plugin-related widgets */
	GList *node;
	
	widget_list = glade_xml_get_widget_prefix (dialog->priv->xml, "preferences_");

	node = widget_list;
	while (node) {
		GtkWidget *widget;
		gchar *name;
		
		widget = node->data;
		name = g_strdup (glade_get_widget_name (node->data));
		/* printf ("Adding widget %s\n", name); */
		
		if (strncmp (name, "preferences_",
                     strlen ("preferences_")) == 0)
		{
			const gchar *property = &name[strlen ("preferences_")];
			register_property_from_glade_string (dialog, widget, property);
		}
		
		node = g_list_next (node);
	}
	g_list_free (widget_list);
	g_list_free (node);
}

static void
pcsx_configure_dialog_init (PcsxConfigureDialog *dialog)
{
	GtkWidget *config_vbox;
	GtkWidget *widget;

	gint system_type;
	
	PcsxConfigureDialogPrivate *priv;

	priv = G_TYPE_INSTANCE_GET_PRIVATE (dialog,
					    PCSX_TYPE_CONFIGURE_DIALOG,
					    PcsxConfigureDialogPrivate);

	dialog->priv = priv;
	
	/* Initialise private variables */

	/* Build the UI and connect signals here */
	priv->xml = glade_xml_new (PACKAGE_DATA_DIR "pcsx_config.glade", "dialog-vbox8", GETTEXT_PACKAGE);
	if (!priv->xml) {
		printf ("Could not open Glade file %s", PACKAGE_DATA_DIR "pcsx_config.glade");
		return;
	}

	/* Get the dialog contents */
	config_vbox = glade_xml_get_widget (priv->xml, "dialog-vbox8");

	/* Add our dialog contents to the vbox of the dialog class */
	gtk_container_add (GTK_CONTAINER (GTK_DIALOG (dialog)->vbox), config_vbox);
	
	/* Set resizable to FALSE so that closing the expander shrinks the dialog */
	gtk_window_set_resizable (GTK_WINDOW (dialog), FALSE);
	
	gtk_dialog_set_has_separator (GTK_DIALOG (dialog), FALSE);
	
	/* Set default widget to Close button */
	widget = gtk_dialog_add_button (GTK_DIALOG (dialog), GTK_STOCK_CLOSE, GTK_RESPONSE_CLOSE);
	gtk_widget_grab_default (widget);
	
	/* Process other widgets from Glade file. Do them before the preferences-related
	   widgets so state-sensitive changes are made later */
	widget = glade_xml_get_widget (priv->xml, "preferences_toggle:1:cpu-psxauto");
	g_signal_connect (G_OBJECT (widget), "toggled",
					  G_CALLBACK (on_autodetect_system_toggled), dialog);
	
	g_object_get (pcsx_gui_prefs, "cpu-psxtype", &system_type, NULL);
	priv->system_type_combo = glade_xml_get_widget (priv->xml, "system_type_combo");
	gtk_combo_box_set_active (GTK_COMBO_BOX (priv->system_type_combo), system_type);
	g_signal_connect (G_OBJECT (priv->system_type_combo), "changed",
					  G_CALLBACK (on_system_type_combo_changed), dialog);
	
	dialog->priv->bios_combo = glade_xml_get_widget (dialog->priv->xml, "preferences_bioscombo:BIOS:bios-bios");
	
	/* Get all the preferences-related widgets and register them (i.e. set their
	   current value and associated callback */
	register_preferences_widgets_from_glade (dialog);

}

GtkWidget *
pcsx_configure_dialog_new (GtkWindow *parent)
{
	GtkWidget *dialog;

	dialog = g_object_new (PCSX_TYPE_CONFIGURE_DIALOG,
						   "title", _("Configure PCSX"),
						   NULL);

	if (parent)
		gtk_window_set_transient_for (GTK_WINDOW (dialog), parent);

	return dialog;

}

static void
pcsx_configure_dialog_response (GtkDialog *dialog, gint response)
{
	switch (response)
	{
		case GTK_RESPONSE_CLOSE:
			/* Close button clicked */
			gtk_widget_destroy (GTK_WIDGET (dialog));

			break;
		case GTK_RESPONSE_DELETE_EVENT:
			/* Dialog closed */
			gtk_widget_destroy (GTK_WIDGET (dialog));

			break;
		default:
			g_assert_not_reached ();
	}
}

static void
pcsx_configure_dialog_destroy (GtkObject *object)
{
	PcsxConfigureDialog *dlg;
	
printf ("Destroying pcsx_configure_dialog...\n");	
	dlg = PCSX_CONFIGURE_DIALOG (object);
	
	if (dlg->priv->xml)
		g_object_unref (dlg->priv->xml);
	
	g_object_unref (dlg->priv);
	
/*	GTK_OBJECT_CLASS (pcsx_configure_dialog_parent_class)->destroy (object);*/
	
printf ("Destroying pcsx_configure_dialog... done\n");
}