/*  Pcsx - Pc Psx Emulator
 *  Copyright (C) 1999-2002  Pcsx Team
 *  Copyright (C) 2009 Andrew Burton
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include <dirent.h>
#include <errno.h>

#include "Linux.h"
#include "Config.h"
#include "plugin-file-handler.h"

/* FIXME */
#include "../libpcsxcore/plugins.h"

PSEgetLibType		PSE_getLibType = NULL;
PSEgetLibVersion	PSE_getLibVersion = NULL;
PSEgetLibName		PSE_getLibName = NULL;

/*
   For a specified plugin type, check that the configured plugin is available.

   Parameters: plugin - e.g. plugin-gpu
*/
static int
plugin_is_available (gchar *plugin) {
	int retval;
	gchar *pluginfile;
	gchar *plugins_dir;
	
	g_return_val_if_fail (plugin != NULL, FALSE);

	retval = FALSE;
	
	g_object_get (pcsx_gui_prefs, "dir-plugins", &plugins_dir, NULL);
	pluginfile = g_build_filename (plugins_dir, plugin, NULL);
	
	g_return_val_if_fail (pluginfile != NULL, FALSE);
	
	printf("Checking if plugin at %s is available\n", pluginfile);
	
	if (g_file_test (pluginfile, G_FILE_TEST_EXISTS))
		retval = TRUE;
	
	g_free (pluginfile);
	g_free (plugins_dir);

	return retval;
}

/* Check the plugins in the .ini file to determine whether they are
   available.
 */
int plugins_configured () {

	gchar *plugin_gpu, *plugin_spu, *plugin_cdr, *plugin_pad1, *plugin_pad2;
	int ret = TRUE;
	
	g_object_get (pcsx_gui_prefs,
				  "plugin-gpu", &plugin_gpu,
				  "plugin-spu", &plugin_spu,
				  "plugin-cdr", &plugin_cdr,
				  "plugin-pad1", &plugin_pad1,
				  "plugin-pad2", &plugin_pad2,
				  NULL);
	
	if (plugin_is_available (plugin_gpu) == FALSE) { Config.Gpu[0] = '\0'; ret = FALSE; }
	if (plugin_is_available (plugin_spu) == FALSE) { Config.Spu[0] = '\0'; ret = FALSE; }
	if (plugin_is_available (plugin_cdr) == FALSE) { Config.Cdr[0] = '\0'; ret = FALSE; }
	if (plugin_is_available (plugin_pad1) == FALSE) { Config.Pad1[0] = '\0'; ret = FALSE; }
	if (plugin_is_available (plugin_pad2) == FALSE) { Config.Pad2[0] = '\0'; ret = FALSE; }
	
	g_free (plugin_gpu);
	g_free (plugin_spu);
	g_free (plugin_cdr);
	g_free (plugin_pad1);
	g_free (plugin_pad2);

	/* If everything is happy, return TRUE */
	return ret;
}

/* For a specified directory scandir, check for all cfg or shared libs and
   symlink to a writeable location. This assumes scandir is unwriteable */
void plugin_file_handler_symlink_plugins (gchar* scandir) {

	GFile *dir;
	GFileEnumerator *enumerator;
	GFileInfo *fileinfo;
	GError *error = NULL;

	gchar *dir_symlink;
	gchar *linkname;
	gchar *cfglinkname;
	gchar *filename;
	int result;

	/* Any plugins found will be symlinked to the following directory */
	dir = g_file_new_for_path (scandir);

	if (dir == NULL) {
		printf(_("Could not open plugins directory: '%s'\n"), scandir);
		/* TODO Should throw an error message and a file chooser for a new plugin directory */
	} else {
		printf("Scanning %s for plugins\n", scandir);

		enumerator = g_file_enumerate_children (dir,
												G_FILE_ATTRIBUTE_STANDARD_NAME,
												G_FILE_QUERY_INFO_NONE,
												NULL,
												&error);
		fileinfo = g_file_enumerator_next_file (enumerator, NULL, &error);

		while (fileinfo != NULL) {
			printf ("  Reading file %s\n", g_file_info_get_name (fileinfo));

			/* If the file name ends with .so (i.e. a shared library) or
			   starts with cfg (i.e. a plugin config utility), symlink it to
			   a readable location */
			if ((g_str_has_prefix (g_file_info_get_name (fileinfo), "cfg")) ||
				(g_str_has_suffix (g_file_info_get_name (fileinfo), ".so"))) {

				printf ("  Symlinking file %s\n", g_file_info_get_name (fileinfo));

				filename = g_build_filename (scandir, g_file_info_get_name (fileinfo), NULL);
		
				/* Generate the full path name of the symlink */
				linkname = g_build_filename (getenv("HOME"), PLUGINS_DIR, g_file_info_get_name (fileinfo), NULL);
				
				/* Create the symlink if it doesnt already exist, or if it points
				   to a different location */
				if (g_file_test (linkname, G_FILE_TEST_EXISTS)) {
					if (g_file_test (linkname, G_FILE_TEST_IS_SYMLINK)) {
						GFile *file;
						GFileInfo *finfo;
						const gchar *target;
						file = g_file_new_for_path (linkname);
						finfo = g_file_query_info (file,
												   G_FILE_ATTRIBUTE_STANDARD_SYMLINK_TARGET,
												   G_FILE_QUERY_INFO_NOFOLLOW_SYMLINKS,
												   NULL,
												   &error);
						target = g_file_info_get_symlink_target (finfo);
						printf ("  Comparing target of existing symlink (%s) to new target (%s)\n",
								target, filename);
						/* If the existing symlink points to a different target, delete it */
						if (g_ascii_strcasecmp (target, filename) != 0) {
							printf ("  Symlink already exists, pointing to %s - will delete\n", target);
							g_file_delete (file, NULL, &error);
							
							/* Symlink file to a readable location */
							result = symlink (filename, linkname);
							if (result != 0) {
								printf ("  Error - %d %s\n", errno, strerror (errno));
							}
						}
					}
				} else {
					/* Symlink file to a readable location */
					result = symlink (filename, linkname);
					if (result != 0) {
						printf ("  Error - %d %s\n", errno, strerror (errno));
					}
				}
			}
		
			fileinfo = g_file_enumerator_next_file (enumerator, NULL, &error);
		}

		g_file_enumerator_close (enumerator, NULL, &error);
		g_object_unref (enumerator);

		/* Change to the symlink dir */
		dir_symlink = g_build_filename (getenv("HOME"), PLUGINS_DIR, NULL);
		g_chdir (dir_symlink);
		g_free (dir_symlink);

	}
}

/* Generate a listing of plugins in the currently selected plugins directory */
void scan_plugins_dir ()
{
	gchar *plugin_dir;
	
	DIR *dir;
	struct dirent *ent;
	
	/* Destroy any existing GList */
	plugin_list = NULL;
	
	g_object_get (pcsx_gui_prefs, "dir-plugins", &plugin_dir, NULL);
	
	printf("Opening directory %s\n", plugin_dir);
	
	dir = opendir(plugin_dir);
	
	g_return_if_fail (dir != NULL);
	
	while ((ent = readdir(dir)) != NULL) {
		void *Handle;
		long type,v;
		char name[256];
		gchar *linkname;
		
		linkname = g_build_filename (plugin_dir, ent->d_name, NULL);
		
		/* Only config files */
		g_return_if_fail (linkname != NULL);
		if (strstr(linkname, ".so") == NULL) continue;
		
		printf("Now processing plugin file %s\n", linkname);

		Handle = SysLoadLibrary (linkname);

		if (Handle == NULL) {
			printf("Error opening %s: %s\n", linkname, dlerror());
			continue;
		}
					
		PSE_getLibType = (PSEgetLibType) dlsym(Handle, "PSEgetLibType");
		PSE_getLibName = (PSEgetLibName) dlsym(Handle, "PSEgetLibName");
		PSE_getLibVersion = (PSEgetLibVersion) dlsym(Handle, "PSEgetLibVersion");
		if (dlerror() == NULL) {
			char ver[32];

			v = PSE_getLibVersion();
			sprintf(ver, " %ld.%ld.%ld",v>>16,(v>>8)&0xff,v&0xff);
			strcat(name, ver);
		}

		pcsx_plugin *plugin;
		plugin = g_new0 (pcsx_plugin, 1);
		plugin->location = g_build_filename (linkname, NULL);
		plugin->type = PSE_getLibType();
		plugin->name = g_strdup (PSE_getLibName());
		plugin->so_name = g_strdup (ent->d_name);
		/*printf("Plugin details: %s - %s - %s - %d\n",
				 plugin->name,
				 plugin->so_name,
				 plugin->location,
				 plugin->type);*/
		
		plugin_list = g_list_append (plugin_list, plugin);
        
		/* Close the handle */
		SysCloseLibrary (Handle);
		Handle = NULL;
	
		g_free (linkname);
	}
	g_free (plugin_dir);

}